import java.util.*;
import java.io.*;

class MixtureVariable {
    double m[];
    int N[];
    
    MixtureVariable(int r) {
	N = new int[r];
	m = new double[r];
    }    
    
    public String toString() {
	StringBuffer b = new StringBuffer(80);
//  	for(int j = 0; j < N.length; j++) {
//  	    if(j > 0)
//  		b.append(" ");
//  	    b.append(N[j]);
//  	}
	for(int i = 0; i < m.length; i++) {
	    if( i > 0 )
		b.append(" ");
	    b.append(Math.floor(m[i]*1000)/1000);
	}
	return b.toString();
    }
}



class MixtureSimulator {

    double p[][], psi[], x[], xtotal;
    MixtureVariable catalysts[], all_chains[];
    double cat_mstar[][], xis[];
    public int num_data, num_components;
    boolean fully_coupled, accept_flag, coalesced;
    // my own random number generator
    MyRandom r;
    // other objects
    FileReader fr;
    BufferedReader br;
    String component_descriptions[];

    // constructor
    public MixtureSimulator(String datafile, boolean first_make_observations) {
	// set up some useful objects
	r = new MyRandom();

	try {
	    // open and read some parameters
	    readParameters(datafile);
	    
	    // before simulating from the posterior, see if
	    // we need to generate some data
	    if( first_make_observations )
		makeObservations(datafile);
	    
	    // read the actual observations
	    readObservations();
	} catch (IOException e) {
	    System.out.println("Error while accessing mixture.obs");
	    System.exit(1);
	}
	
	// and set up some extra variables used by the gibbs sweeper
	psi = new double[num_data];
	x = new double[num_data+num_components];

    }
    
    // generate a Dirichlet(N[1]+1,...,N[r]+1) into the 
    // array m[]. 
    void dirichlet(double m[], int N[], double x[]) {
	double s = 0;
	
	if( x == null ) {
	    for(int k = 0; k < num_components; k++) {
		if( N == null )
		    m[k] = r.nextExponential(1);
		else {
		    m[k] = 0;
		    for(int j = 0; j < N[k]+1; j++ )
			m[k] += r.nextExponential(1);
		}
		s += m[k];
	    }
	} else { // assume N != null
	    int a = 0;
	    for(int k = 0; k < num_components; k++) {
		m[k] = 0;
		for(int j = 0; j < N[k]+1; j++ ) {
		    m[k] += x[a];
		    a++;
		}
		s += m[k];
	    }
	}
	for(int c = 0; c < num_components; c++) 
	    m[c] /= s;
    }

    // this calculates the N_k function for a given m
    void precalc_N(int N[], double m[], double psi[]) {
	for(int s = 0; s < num_data; s++) {
	    // calculate the probability distribution
	    // proportional to m_k p_k(eta_s)
	    double sums[] = new double[num_components];
	    double sum = 0;
	    for(int k = 0; k < num_components; k++) {
		sum += m[k]*p[k][s];
		sums[k] = sum;
	    }
	    if( psi == null )
		sum *= r.nextDouble();
	    else
		sum *= psi[s]; 
	    // calculate the functions N_k(Z)
	    for(int k = 0; k < num_components; k++) {
		if( sum <= sums[k] ) {
		    N[k]++;
		    break;
		}
	    }
	}
    }

    public MixtureVariable bigSweep(MixtureVariable ref_chain, int num_sweeps,
				 int cats_per_sweep, int num_test_points) {
	// generate the basic update
	initCoupledGibbsSweep(cats_per_sweep);

	if( cats_per_sweep <= 0 ) {
	    fully_coupled = false; // certainly
	    coalesced = false;
	} else { // cats_per_sweep > 0
	    coalesced = true; // be optimistic
	    fully_coupled = true; // perhaps
//  	    for(int z = 0; z < cats_per_sweep; z++)
//  	    System.out.println("cat_"+z+"="+catalysts[z]+" xi="+xis[z]
//  			       +" mstar="
//  			       +(Math.floor(1000*cat_mstar[z][0])/1000)+" "
//  		+(Math.floor(1000*cat_mstar[z][1])/1000));

	    // test to see if fully_coupled
	    for(int t = 0; t < num_test_points; t++) {
		MixtureVariable test = new MixtureVariable(num_components);
		// pick a point uniformly in S
		dirichlet(test.m, null, null);
		// work out where it's headed under the basic update
		precalc_N(test.N, test.m, psi);
		
		// done! We now have the basic move 
		// next we cycle through the catalysts
		boolean accepted = false;
		for(int c = 0; c < catalysts.length; c++) {
		    // calculate the acceptance ratio
		    double logratio = 0;
		    for(int k = 0; k < num_components; k++) {
			logratio += (catalysts[c].N[k] - test.N[k]) * 
			    (Math.log(test.m[k]/cat_mstar[c][k]));
		    }
//  		    System.out.println(" logratio="+logratio);
		    if( Math.exp(logratio) > xis[c] ) {// accepts the move
			accepted = true;
			break; // cycle
		    }
		}
		if( !accepted ) { // the test point failed to be caught
		    fully_coupled = false;
		    break; // test cycle
		}
	    } // test cycle
	}
	
	if( fully_coupled ) { // ok, we can propagate all the various chains
	    // so create a list of all the chains we must track
	    all_chains = new MixtureVariable[cats_per_sweep+1];
	    all_chains[0] = evalCoupledGibbsSweep(ref_chain);
	    System.arraycopy(catalysts, 0, all_chains, 1, cats_per_sweep);
	    // there are (num_sweeps - 1) extra sweeps to go
	    for( int sweep = 1; sweep < num_sweeps; sweep++) {
		initCoupledGibbsSweep(cats_per_sweep);
		for( int c = 0; c < all_chains.length; c++) {
		    all_chains[c] = evalCoupledGibbsSweep(all_chains[c]);
		}
	    }
	    // the day of reckoning: see if we had coalescence
	    // by comparing values with the reference chain
	    away: {
		for( int c1 = 1; c1 < all_chains.length; c1++) {
		    for(int l = 0; l < num_components; l++)
			if( all_chains[c1].m[l] != all_chains[0].m[l] ) {
			    coalesced = false;
			    break away;
			}
		}
	    }
	    return all_chains[0]; // this is the propagated value of refChain
	} else { // not fully_coupled
	    coalesced = false;
	    // don't bother with extra chains
	    MixtureVariable propagated;
	    // do one obligatory sweep
	    propagated = evalCoupledGibbsSweep(ref_chain);
	    // and then the others
	    for( int sweep = 1; sweep < num_sweeps; sweep++) {
		initCoupledGibbsSweep(cats_per_sweep);
		propagated = evalCoupledGibbsSweep(propagated);
	    }
	    return propagated;
	}
    }

    
    void initCoupledGibbsSweep(int num_catalysts) {
	// begin by generating the random variables needed for
	// an ordinary Gibbs sweep
	for(int i = 0; i < num_data; i++) {
	    psi[i] = r.nextDouble();
	}
	for(int i = 0; i < num_data+num_components; i++) {
	    x[i] = r.nextExponential(1);
	}

	// next we generate all the catalysts
	catalysts = new MixtureVariable[num_catalysts];
	xis = new double[num_catalysts];
	cat_mstar = new double[num_catalysts][num_components];
	for(int i = 0; i < num_catalysts; i++) {	    
	    // select m_star uniformly Dirichlet(1,1,...,1)
	    dirichlet(cat_mstar[i], null, null);

	    // now generate the catalyst
	    catalysts[i] = new MixtureVariable(num_components);
	    // begin by precalculating the functions N_k(catalyst)
	    precalc_N(catalysts[i].N, cat_mstar[i],null);
	    // next we generate a Dirichlet(N_1+1,...,N_r+r)
	    dirichlet(catalysts[i].m, catalysts[i].N, null);
	    
	    // and generate a test uniform to accept/reject catalyst proposal
	    xis[i] = r.nextDouble();
	}
    }

    MixtureVariable evalCoupledGibbsSweep(MixtureVariable from) {
	MixtureVariable to = new MixtureVariable(num_components);
	accept_flag = false;

	// first evaluate the ordinary Gibbs sweep
	precalc_N(to.N, from.m, psi);
	// next we generate a Dirichlet(N_1+1,...,N_r+1)
	dirichlet(to.m, to.N,x);
	
	// done! We now have the basic move 
	// next we cycle through the catalysts
	for(int i = 0; i < catalysts.length; i++) {
	    // calculate the acceptance ratio
	    double logratio = 0;
	    for(int k = 0; k < num_components - 1; k++) {
		logratio += (catalysts[i].N[k] - to.N[k]) * 
		    (Math.log(from.m[k]/cat_mstar[i][k]));
	    }
	    if( Math.exp(logratio) > xis[i] ) {// accepts the move
		System.arraycopy(catalysts[i].m, 0, to.m, 0, num_components);
		System.arraycopy(catalysts[i].N, 0, to.N, 0, num_components);
		accept_flag = true;
	    }
	}
	return to;
    }


    // this opens the file and figures out how many components we have
    public void readParameters(String file) throws IOException {
	// open the file
        fr = new FileReader(file);
        br = new BufferedReader(fr,8192);
	String line = br.readLine();
	if( !line.startsWith("#h mixture") ) {
	    System.out.println("The file mixture.obs has the wrong format!");
	    fr.close();
	    System.exit(1);
	}
	// expect the first line in the format
	// #h mixture has 5 components and 30 datapoints
	StringTokenizer st = new StringTokenizer(line);
	st.nextToken(); // skip #h
	st.nextToken(); // skip mixture
	st.nextToken(); // skip has
	try {
	    num_components = Integer.parseInt(st.nextToken());
	} catch (NumberFormatException e) {
	    System.out.println("Can't read the number of components");
	    fr.close();
	    System.exit(1);
	} 
	st.nextToken(); // skip components
	st.nextToken(); // skip and
	try {
	    num_data = Integer.parseInt(st.nextToken());
	} catch (NumberFormatException e) {
	    System.out.println("Can't read the number of datapoints");
	    fr.close();
	    System.exit(1);
	} 

	System.out.println("Mixture has " + num_components +
			   " components and " + num_data + " datapoints.");

	// read in the components
	component_descriptions = new String[num_components];
	for(int i = 0; i < num_components; i++) {
	    line = br.readLine();
	    if( !line.startsWith("#") ) {
		System.out.println("Error reading component "+(i+1));
		System.exit(1);
	    }
	    component_descriptions[i] = line;
	}
	// close the file reader somewhere else
    }

    public void makeObservations(String file) throws IOException {
	fr.close(); // we're going to replace the file

	FileWriter fw = new FileWriter(file);
	BufferedWriter bw = new BufferedWriter(fw, 8192);
	// recreate the header
	fw.write("#h mixture has " + num_components + " components and "
		 + num_data + " datapoints.\n");
	for(int i = 0; i < num_components; i++) {
	    fw.write(component_descriptions[i] + "\n");
	}
	// read in the mixture description
	double m[] = new double[num_components];
	double param1[] = new double[num_components];
	double param2[] = new double[num_components];

	for(int i = 0; i < num_components; i++) {
	    StringTokenizer st = new StringTokenizer(component_descriptions[i], "% ");
	    st.nextToken(); // skip #
	    try{ 
		m[i] = Double.valueOf(st.nextToken()).doubleValue();
		if( i > 0 )
		    m[i] += m[i-1];
	    } catch (NumberFormatException e) {
		System.out.println("Error processing component #" + (i+1));
		fw.close();
		System.exit(1);
	    }
	    st.nextToken(); // skip normal
	    try{ 
		param1[i] = Double.valueOf(st.nextToken()).doubleValue();
	    } catch (NumberFormatException e) {
		System.out.println("Error processing component #" + (i+1));
		fw.close();
		System.exit(1);
	    }
	    try{ 
		param2[i] = Double.valueOf(st.nextToken()).doubleValue();
	    } catch (NumberFormatException e) {
		System.out.println("Error processing component #" + (i+1));
		fw.close();
		System.exit(1);
	    }
	}

	System.out.println("Generating new data from the mixture");

	// now make up some new data
	for(int k = 0; k < num_data; k++) {
	    double xi = 100*r.nextDouble();
	    for(int j = 0; j < num_components; j++) {
		if( xi <= m[j] ) {
		    double x = param1[j] + 
			Math.sqrt(param2[j])* r.nextGaussian();
		    bw.write(""+x+"\n");
		    break;
		}	
	    }	    
	}
	bw.close(); // done!

	// now reopen the file for reading and skip to the data section
	fr = new FileReader(file);
	br = new BufferedReader(fr, 8192);
	for(int i = 0; i < num_components+1; i++) {
	    String line = br.readLine();
	}
	// data lines should start now
    }

    // this reads the observations and calculates the likelihood table p
    public void readObservations() throws IOException {
	// assume the file is open and ready to read
	double datapoints[] = new double[num_data];
	for(int i = 0; i < num_data; i++) {
	    String line = br.readLine();
	    if( line == null ) {
		System.out.println("No data!");
		fr.close();
		System.exit(1);
	    }
	    try {
		datapoints[i] = Double.valueOf(line).doubleValue();
	    } catch (NumberFormatException e) {
		System.out.println("Error reading datapoint #"+(i+1));
		fr.close();
		System.exit(1);
	    }
	}
	fr.close();
	System.out.println("Initializing MixtureSimulator");
	// now create a big table of likelihoods
	p = new double[num_components][num_data];
	for(int i = 0; i < num_components; i++) {
	    StringTokenizer st = new StringTokenizer(component_descriptions[i]);
	    st.nextToken(); // skip #
	    st.nextToken(); // skip 23%
	    st.nextToken(); // skip normal
	    try {
		double mu = Double.valueOf(st.nextToken()).doubleValue();
		double sigma2 = Double.valueOf(st.nextToken()).doubleValue();
		// compute some likelihoods
		for(int j = 0; j < num_data; j++) {
		    double w = datapoints[j] - mu;
		    p[i][j] = Math.exp( - (w*w)/(2*sigma2));
		}
	    } catch (NumberFormatException e) {
		System.out.println("Error processing component #"+(i+1));
		System.exit(1);
	    }
	}
    }


    // this is the executable block
    public static void main(String args[]) throws IOException {
	// some program defaults
	String datafile = "mixture.obs";
	int number_of_sweeps = 20;
	int number_of_catalysts_per_sweep = 20;
	int number_of_iterations = 50;
	int number_of_test_points = 50;
	boolean verbose = false;
	boolean first_make_observations = false;
	boolean force_output = false;

	// read in the command line arguments
	for( int i = 0; i < args.length; i++) {
	    if( args[i].charAt(0) != '-' ) {
		System.out.println("Unrecognized command line switch");
	    }
	    switch( args[i].charAt(1) ) {
	    case 'v':
	    case 'V': // show the user what's going on
		verbose = true;
		break;
	    case 's':
	    case 'S': // number of sweeps
		try {
		    number_of_sweeps = Integer.parseInt(args[++i]);
		} catch (NumberFormatException e){
		   System.out.println("Unrecognized command line switch");
		   number_of_sweeps = 20;
		}
		break;
	    case 'c': // number of catalysts per sweep
	    case 'C':
		try {
		    number_of_catalysts_per_sweep = Integer.parseInt(args[++i]);
		} catch (NumberFormatException e){
		    System.out.println("Unrecognized command line switch");
		    number_of_catalysts_per_sweep = 20;
		}
		break;
	    case 'i': // number of iterations
	    case 'I':
		try {
		    number_of_iterations = Integer.parseInt(args[++i]);
		} catch (NumberFormatException e){
		    System.out.println("Unrecognized command line switch");
		    number_of_iterations = 50;
		}
		break;
	    case 'f':
	    case 'F': // force everything to be written in mixture.dat
		force_output = true;
		break;
	    case 'o': 
	    case 'O': // create new observations first (placed in datafile)
		first_make_observations = true;
		break;
	    case 't':
	    case 'T': // number of test points to use
		try {
		    number_of_test_points = Integer.parseInt(args[++i]);
		} catch (NumberFormatException e){
		    System.out.println("Unrecognized command line switch");
		    number_of_test_points = 50;
		}			
	    }
	}

	MixtureSimulator ms = new MixtureSimulator(datafile,  
						   first_make_observations);

	System.out.println("MixtureSimulator is running");
    
	// initialize the chain uniformly.
        MixtureVariable ref_chain = new MixtureVariable(ms.num_components);
	for(int j=0; j < ms.num_components; j++) {
	    ref_chain.m[j] = 1.0/ms.num_components; // say
	}
	if( verbose ) {
	    System.out.println("Number of iterations: " + number_of_iterations
			       + " using " + number_of_sweeps 
			       + " sweeps/iteration and " +
			       number_of_catalysts_per_sweep +
			       " catalysts/sweep.");
	    System.out.println(" X_0 = ["+ref_chain+"]");
	}

	// open the output file
	FileWriter fw = new FileWriter("mixture.dat");
	BufferedWriter f = new BufferedWriter(fw, 8192);
	f.write("#h");
	for(int i = 0; i < ms.num_components; i++) {
	    f.write(" m_"+(i+1));
	}
	if( force_output )
	    f.write(" forced output NOT IID!\n");
	else
	    f.write("\n");

	boolean first = true;
	int j = 1;
	MixtureVariable old_chain = new MixtureVariable(ms.num_components);
	for(int i = 0; i < number_of_iterations; i++) {
	    // save the state just prior to testing for coalescence
	    System.arraycopy(ref_chain.m,0,old_chain.m,0,ms.num_components);
	    // do the coalescence thing
	    ref_chain = ms.bigSweep(ref_chain, number_of_sweeps, 
				    number_of_catalysts_per_sweep,
				    number_of_test_points);
	    if( verbose ) {
		System.out.print((i+1)+":");
		if( ms.fully_coupled )
		    System.out.print(" fully coupled");
		else
		    System.out.print(" not coupled");
		if( ms.coalesced ) {
		    System.out.println(" + coalesced");
		    System.out.print(" T_"+j+" = "+i);
		} else
		    System.out.println(" - not coalesced");

		System.out.println(" X_"+(i+1)+" = ["+ref_chain+"]");
	    }

	    if( force_output ) {
		f.write(old_chain.toString());
		f.write("\n");
		if( ms.coalesced )
		    j++;
	    } else if( ms.coalesced ) { 
		j++;
		if( first ) {
		    first = false;
		} else { // write the output to the file
		    f.write(old_chain.toString());
		    f.write("\n");
		}
	    } 
	}

	System.out.println("Produced an IID sample of size " + (j-2) );
	// close the file
	f.close();
    }
    
}














